/*****************************************************************************
 * $LastChangedDate: 2009-05-18 13:53:36 -0500 (Mon, 18 May 2009) $
 * @file
 * @author	Jim E. Brooks  http://www.palomino3d.org
 * @brief   Template class for sharing objects by pointer.
 * @remarks Documented at shared_ptr_doc.hh.
 *//*
 * LEGAL:	COPYRIGHT (C) JIM E. BROOKS 2007
 *          THIS SOURCE CODE IS RELEASED UNDER THE TERMS
 *          OF THE GNU GENERAL PUBLIC LICENSE VERSION 2 (GPL 2).
 *****************************************************************************/

// (re-included by shared_ptr.hh)

////////////////////////////////////////////////////////////////////////////////
// SharedPtr
//
// - Shared constructs a new object with initially mRefCnt=0.
//   Object* pObj = new Object;       // mRefCnt=0
//   SharedPtr<Object> shObj = pObj;  // mRefCnt=1
//   Therefore, mRefCnt=0 is NOT an accurate indication of a destroyed Shared.
// - SHPTR_CHECK() can catch reviving a destroyed Shared by checking its typesig.
//
template<typename SHARED>
class SharedPtr
{
COPYABLE(SharedPtr)
template<typename SHARED_PTR> friend class SharedPtr;
private:
#if DEBUG
	template<typename SHARED_PTR>
	void SHPTR_CHECK( const SHARED_PTR* const sharedPtr ) const
	{
		// sharedPtr never should be NULL, but mObj may be NULL.
		// SharedPtr should point to a Shared with minimum of 1 ref-cnt.
		CHECK_TYPESIG(sharedPtr,TYPESIG_SHARED_PTR);
		IF_NOT_NULL( sharedPtr->mObj )
		{
			CHECK_TYPESIG(sharedPtr->mObj,TYPESIG_SHARED);
			ASSERT( sharedPtr->mObj->mRefCnt > 0 );
		}
	}
	void SHPTR_CHECK( const SHARED* const shared ) const
	{
		// Shared initially has zero references.
		// A case where handling NULL is unavoidable.
		IF_NOT_NULL( shared )
		{
			CHECK_TYPESIG(shared,TYPESIG_SHARED);
			ASSERT( shared->mRefCnt >= 0 );
		}
	}
#else
#	undef SHPTR_CHECK
#	define SHPTR_CHECK(SHR)
#endif

public:

	// Default ctor necessary to use SharedPtrs in STL containers.
	SharedPtr( void )
	{
	SET_TYPESIG(this,TYPESIG_SHARED_PTR);
#if SHARED_PTR_NULL
		mObj = NULL;        // hidden landmine
#else
		mObj = new SHARED;  // need a real Shared object for non-null SharedPtr class
		mObj->mRefCnt = 1;  // this SharedPtr is the first to reference object
#endif
	}

    SharedPtr( SHARED* obj )
    {
	SET_TYPESIG(this,TYPESIG_SHARED_PTR);
	SHPTR_CHECK(obj);
        mObj = obj;
		IF_NOT_NULL(obj) ++(obj->mRefCnt);  // passed obj may have other references
    }

    SharedPtr( const SharedPtr& src )
    {
	// The compiler won't choose the following templatized ctor as a copy ctor
	// (though it is suited) so this non-template copy ctor remains necessary.

	SET_TYPESIG(this,TYPESIG_SHARED_PTR);
	SHPTR_CHECK(&src);
        mObj = src.mObj;
		IF_NOT_NULL(src.mObj)
		{
			++(src.mObj->mRefCnt);  // copying a SharedPtr increments the ref-cnt
		}
    }

	template<typename SHARED2>
    SharedPtr( const SharedPtr<SHARED2>& src )
    {
	// -- Same as SharedPtr( const SharedPtr& src ) --
	// This allows passing SharedPtr<Derived> as SharedPtr<Base>.

	SET_TYPESIG(this,TYPESIG_SHARED_PTR);
	SHPTR_CHECK(&src);
        mObj = src.mObj;
		IF_NOT_NULL(src.mObj)
		{
			++(src.mObj->mRefCnt);  // copying a SharedPtr increments the ref-cnt
		}
    }

    ~SharedPtr()
    {
	CHECK_TYPESIG(this,TYPESIG_SHARED_PTR);
		IF_NOT_NULL(mObj)
		{
	    SHPTR_CHECK(this);
			int refCnt = --(mObj->mRefCnt);
			if ( refCnt == 0 )  // delete object if last reference
			{
				delete mObj;
				DEBUG_CODE( mObj = NULL; )
			}
		}
	INVALIDATE_TYPESIG(this,TYPESIG_SHARED_PTR);
    }

    // Reassign this SharedPtr with a different object.
    SharedPtr& operator=( SHARED* obj )
    {
	SHPTR_CHECK(this);
	SHPTR_CHECK(obj);
        IF_NOT_NULL(obj)
		{
			++(obj->mRefCnt);  // order-dependent to handle self-assignment
		}
		IF_NOT_NULL(mObj)
		{
	        if ( --(mObj->mRefCnt) == 0 )  // decrease outgoing object's ref-cnt
				delete mObj;
		}
        mObj = obj;
        return *this;
    }

    SharedPtr& operator=( const SharedPtr& src )
    {
	SHPTR_CHECK(this);
	SHPTR_CHECK(&src);
        // Increment ref-cnt of the src SharedPtr (copy operation).
        // Decrement ref-cnt of this SharedPtr before overwriting it.
		IF_NOT_NULL(src.mObj)
		{
	        ++(src.mObj->mRefCnt);  // order-dependent to handle self-assignment
		}
		IF_NOT_NULL(mObj)
		{
	        if ( --(mObj->mRefCnt) == 0 )
				delete mObj;
		}
        mObj = src.mObj;
        return *this;
    }

    SHARED* operator->( void ) const { SHPTR_CHECK(this); return mObj; }

    // Limited use (extracted reference could dangle).
    SHARED& operator*( void ) const { SHPTR_CHECK(this); return *mObj; }

    bool operator==( const SharedPtr& other ) const { SHPTR_CHECK(this); SHPTR_CHECK(&other); return mObj == other.mObj; }
    bool operator!=( const SharedPtr& other ) const { SHPTR_CHECK(this); SHPTR_CHECK(&other); return mObj != other.mObj; }
    bool operator==( const SHARED* obj )      const { SHPTR_CHECK(this); SHPTR_CHECK(obj);    return mObj == obj; }
    bool operator!=( const SHARED* obj )      const { SHPTR_CHECK(this); SHPTR_CHECK(obj);    return mObj != obj; }

    // Limited use.
	SHARED*             PTR( void ) const { SHPTR_CHECK(this); return mObj; }
	const SHARED* CONST_PTR( void ) const { SHPTR_CHECK(this); return mObj; }
	SHARED&             REF( void ) const { SHPTR_CHECK(this); return *mObj; }
	const SHARED& CONST_REF( void ) const { SHPTR_CHECK(this); return *mObj; }
	uint          GetRefCnt( void ) const { SHPTR_CHECK(this); return mObj->GetRefCnt(); }

private:
	SHARED*		mObj;
public:
	DECL_TYPESIG(TYPESIG_SHARED_PTR)
};

template<typename T> bool operator==( const SharedPtr<T>& a, const SharedPtr<T>& b ) { return a.PTR() == b.PTR(); }
template<typename T> bool operator!=( const SharedPtr<T>& a, const SharedPtr<T>& b ) { return a.PTR() != b.PTR(); }
template<typename T> bool operator<( const SharedPtr<T>& a, const SharedPtr<T>& b ) { return a.PTR() < b.PTR(); }

#undef SHPTR_CHECK
