/******************************************************************************
 * $LastChangedDate: 2009-05-18 09:11:03 -0500 (Mon, 18 May 2009) $
 * @file
 * @author	Jim E. Brooks  http://www.palomino3d.org
 * @brief	Streams.
 *//*
 * LEGAL:	COPYRIGHT (C) 2004 JIM E. BROOKS
 *          THIS SOURCE CODE IS RELEASED UNDER THE TERMS
 *          OF THE GNU GENERAL PUBLIC LICENSE VERSION 2 (GPL 2).
 ******************************************************************************/

#define BASE_STREAM_CC 1
#include "base/module.hh"
#include "base/stream.hh"
#include "base/file.hh"
#undef cout
#undef cerr

namespace base {

////////////////////////////////////////////////////////////////////////////////
/////////////////////////////  functions  //////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

/*****************************************************************************
 * Open a stream to a file.
 * @param	fs
 *			File stream.
 * @param	bits
 *			Open for reading and/or writing.
 * @param	filename
 * @return	False if stream error.
 *****************************************************************************/
bool
StreamOpen( fstream& fs, ios::openmode bits, const string& filename )
{
	// Binary else trouble.
	bits |= ios::binary;

	// Try opening a stream to a file.
	// For the open-for-writing case, first try to open the file without the truncate bit.
	// But some iostream implementations might fail if the file doesn't exist.
	fs.open( filename.c_str(), bits );
	if ( (not fs.good())				// if error
     and (bits & ios::out)				// if output stream
     and (not (bits & ios::trunc)) )	// if truncate not already set
	{
		// This might be the file-doesn't-exist case depending
		// on the iostream implementation.  Retry with truncate bit.
		fs.clear();
		fs.open( filename.c_str(), bits | ios::trunc );
	}

	return fs.good();
}

/*****************************************************************************
 * Read all chars to EOL.
 * @return False if stream error.
 *****************************************************************************/
bool
StreamReadLine( istream& strm, string& line )
{
	line.erase();
	char c;
	while ( strm.good() and (c = strm.get()) != '\n' )
	{
		line += c;
	}
	return strm.good();
}

/*****************************************************************************
 * Jump to the next line.
 * @return False if stream error.
 *****************************************************************************/
bool
StreamToNextLine( istream& strm )
{
	while ( strm.good() && (strm.get() != '\n') ) { }
	return strm.good();
}

/*****************************************************************************
 * Peek input stream.
 * @return True if word is in input stream.
 *****************************************************************************/
bool
StreamPeek( istream& strm, const string& word )
{
	bool result = false;

	// Save read position.
	std::ios::pos_type pos = strm.tellg();

	try
	{
		// Read char-by-char into a buf
		string buf;
		for ( uint i = 0; i < word.size(); ++i )
		{
			const char c = char(strm.get());
			if ( not strm.good() )
				break;

			switch ( c )
			{
				case CHAR_LF: case CHAR_CR: case ' ': case '\t':
				{
					--i;  // don't append whitespace to buf[]
					continue;
				}
				break;

				default:
				{
					buf += char(c);
				}
				break;
			}

			if ( buf[i] != word[i] )  // mismatch?
				break;
		}
		result = (buf == word) && strm.good();
	}
	catch (...)
	{
		result = false;
	}

	// Restore read position.
	strm.seekg( pos );

	return result;
}

////////////////////////////////////////////////////////////////////////////////
////////////////////////////  StreambufFile2  //////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

/*****************************************************************************
 * ctor/dtor.
 * @param	file0, file1
 *          Second is optional.
 * @param	close0, close1
 *			If destructor should close file.  Never deletes FILE* object.
 *          Second is optional.
 *****************************************************************************/
StreambufFile2::StreambufFile2( SafePtr<FILE> file0, bool close0,
                                SafePtr<FILE> file1, bool close1 )
:	mFile0(file0),
	mFile1(file1),
	mClose0(close0),
	mClose1(close1)
{
}

StreambufFile2::~StreambufFile2()
{
	// First file is always present, second is optional.
	if ( mClose0 )                   fclose( mFile0.PTR() );
	if ( mClose1 && mFile1 != NULL ) fclose( mFile1.PTR() );
}

/*****************************************************************************
 * Either is called by iostream to write out buffer.  xsputn() is optional/optimal.
 *****************************************************************************/
int
StreambufFile2::overflow( int c )
{
                          fputc( c, mFile0.PTR() );
	if ( mFile1 != NULL ) fputc( c, mFile1.PTR() );
	return ~EOF;
}

std::streamsize
StreambufFile2::xsputn( const char* buf, std::streamsize n )
{
    size_t written  = n;  // mainly to shutup gcc 4.3
    size_t written2 = n;
                          written  = fwrite( buf, n, 1, mFile0.PTR() );
	if ( mFile1 != NULL ) written2 = fwrite( buf, n, 1, mFile1.PTR() );
	return written;
}

} // namespace base
