/*****************************************************************************
 * $LastChangedDate: 2009-05-18 11:33:20 -0500 (Mon, 18 May 2009) $
 * @file
 * @author	Jim E. Brooks  http://www.palomino3d.org
 * @brief	Streams.
 *//*
 * LEGAL:	COPYRIGHT (C) 2004 JIM E. BROOKS
 *          THIS SOURCE CODE IS RELEASED UNDER THE TERMS
 *          OF THE GNU GENERAL PUBLIC LICENSE VERSION 2 (GPL 2).
 *****************************************************************************/

#ifndef BASE_STREAM_HH
#define BASE_STREAM_HH 1

#include <iostream>
#include <iomanip>
#include <streambuf>
#include <fstream>

namespace base {

// C++ stream field formatting.
#define OMANIP_FIELD(CHARS) std::setw(CHARS) << std::right

/// Helpers for C++ streams.

//bool StreamGood( std::ios& s );  // just use stream.good() (hi-freq)

bool
StreamOpen( fstream& fs, ios::openmode bits, const string& filename );

bool
StreamReadLine( istream& strm, string& line );

bool
StreamToNextLine( istream& strm );

bool
StreamPeek( istream& strm, const string& word );

////////////////////////////////////////////////////////////////////////////////
/// @brief Streambuf class that can write to one or two files (a confluence/fork).
///
/// This is a "streambuf" class to serve as the basis of an ostream.
/// Derived from Nicolai Josuttis's STL book.
///
/// @verbatim
/// Example:
///		SafePtr<FILE> file1 = fopen(...);
///		SafePtr<FILE> file2 = fopen(...);
///     StreamFile2 streambufFile2( file1, file2 );
///     std::ostream out( &streambufFile2 );  // ostream based on streambuf
///     out << "hello" << endl;
/// @endverbatim
///
class StreambufFile2 : public streambuf
{
PREVENT_COPYING(StreambufFile2)
typedef streambuf Parent;
public:
	StreambufFile2( SafePtr<FILE> file0, bool close0,
                    SafePtr<FILE> file1 = NULL, bool close1 = false );
	~StreambufFile2();

protected:
	virtual int_type	overflow( int_type c );
	std::streamsize		xsputn( const char* buf, std::streamsize n );

	SafePtr<FILE>	mFile0, mFile1;    // stream outputs to both files
	bool			mClose0, mClose1;  // true if destructor should close the file
};

////////////////////////////////////////////////////////////////////////////////
/// @brief NopStream is a indirect way to use a disabled C++ stream.
///
/// This is a "streambuf" class to serve as the basis of an ostream.
/// Derived from Nicolai Josuttis's STL book.
///
/// @verbatim
/// Example:
///     NopStreambuf nopStreamBuf;
///     std::ostream gLog( &gNopStreambuf );
/// @endverbatim
///
class NopStreambuf : public streambuf
{
PREVENT_COPYING(NopStreambuf)
typedef streambuf Parent;
public:
	NopStreambuf( void ) { }
	~NopStreambuf() { }

protected:
	virtual int_type	overflow( int_type c ) { return ~EOF; }
	std::streamsize		xsputn( const char* buf, std::streamsize n ) { return n; }
};

} // namespace base

#endif // BASE_STREAM_HH
