/*****************************************************************************
 * $LastChangedDate: 2009-05-21 11:41:41 -0500 (Thu, 21 May 2009) $
 * @file
 * @author	Jim E. Brooks  http://www.palomino3d.org
 * @brief	Base scene-graph class.
 *//*
 * LEGAL:	COPYRIGHT (C) 2009 JIM E. BROOKS
 *          THIS SOURCE CODE IS RELEASED UNDER THE TERMS
 *          OF THE GNU GENERAL PUBLIC LICENSE VERSION 2 (GPL 2).
 *****************************************************************************/

#define GRAPH_BASE_SCENE_GRAPH_CC 1
#include "base/module.hh"
#include "base/conf.hh"
using namespace base;
#include "gfx/module.hh"
#include "gfx/osg.hh"
#include "gfx/osg_funcs.hh"
using namespace gfx;
#include "object/module.hh"
#include "object/events.hh"
using namespace object;
#include "view/module.hh"
#include "view/node_masks.hh"
using namespace view;
#include "shader/module.hh"
#include "shader/shader_factory.hh"
using namespace shader;
#include "graph/module.hh"
#include "graph/osg.hh"
#include "graph/conf.hh"
#include "graph/base_scene_graph.hh"

#include <osg/BlendFunc>
#include <osgUtil/Optimizer>

namespace graph {

////////////////////////////////////////////////////////////////////////////////
////////////////////////////  BaseSceneGraph  //////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

NodeSort BaseSceneGraph::msBranchNodeSort;  // default

/*****************************************************************************
 * ctor/dtor.
 *****************************************************************************/
BaseSceneGraph::BaseSceneGraph( RefPtr<GroupNode> rootNode )
:	mBaseRootNode(rootNode),
	mBranchNode(msBranchNodeSort.CreateGroupNode()),  // create group with default state
	mNodeSortMap(),
	mObjects()
{
    // Initialize root node.
    SetRootNode( rootNode );

	// Unlisten() isn't necessary for singletons.
    EVENT_DESTROY_OBJECT.Listen( new DestroyObjectListener( *this ) );
}

BaseSceneGraph::~BaseSceneGraph()
{
	// NOP
}

/*****************************************************************************
 * (protected)
 *****************************************************************************/
void
BaseSceneGraph::SetRootNode( RefPtr<GroupNode> rootNode )
{
    // Switch root node.
    RefPtr<GroupNode> prevRootNode = mBaseRootNode;
    mBaseRootNode = rootNode;

	// Attach branch node.
	mBaseRootNode->addChild( mBranchNode.get() );

    // Set default state on root node.
    SetDefaultState( mBaseRootNode.get() );

    // Detach children from outgoing root node.
    if ( (prevRootNode.get() != NULL)
     and (prevRootNode.get() != mBaseRootNode.get()) )
    {
        OsgRemoveChildren( prevRootNode );
    }
}

/*****************************************************************************
 * (protected)
 *****************************************************************************/
void
BaseSceneGraph::SetDefaultState( RefPtr<Node> node )
{
    RefPtr<osg::StateSet> stateSet = node->getOrCreateStateSet();
    const osg::StateAttribute::OverrideValue override = osg::StateAttribute::OVERRIDE;

    // Disable hw lighting.
    stateSet->setMode( GL_LIGHTING, osg::StateAttribute::OFF | override );

	// Set default shader.
	GET_SHADER_FACTORY().AttachShader( ShaderName(), node.get() );

    // Set default blend function (don't enable blending).
    RefPtr<osg::BlendFunc> blendFunc = new osg::BlendFunc;
    blendFunc->setFunction( osg::BlendFunc::SRC_ALPHA, osg::BlendFunc::ONE_MINUS_SRC_ALPHA );
    stateSet->setAttributeAndModes( blendFunc.get(), osg::StateAttribute::ON );

    // Enable back-face culling.
    stateSet->setMode( GL_CULL_FACE, osg::StateAttribute::ON | override );
}

/*****************************************************************************
 * (private subroutine)
 *****************************************************************************/
void
BaseSceneGraph::AttachNode( RefPtr<Node> node, RefPtr<GroupNode> parent )
{
    // Ensure node is unattached.
    if ( EX( node->getNumParents() == 0 ) )
	{
		// Attach node.
		parent->addChild( node.get() );
	}

ASSERT( node->getNumParents() == 1 );
}

/*****************************************************************************
 * Attach node to scene-graph under branch node with special graphical-state.
 *****************************************************************************/
void
BaseSceneGraph::AttachNode( RefPtr<Node> node, const NodeSort& nodeSort )
{
CHECK_TYPESIG(&nodeSort,TYPESIG_NODE_SORT);

	// Optimize this node's subgraph.
	Optimize( node, nodeSort );

    // Try to map NodeSort to a cached group node.
    NodeSortMap::iterator iter = mNodeSortMap.find( nodeSort );
    if ( iter != mNodeSortMap.end() )
    {
    	// Found cached group node.
    	RefPtr<GroupNode> groupNode = iter->second;
    	AttachNode( node, groupNode );
    }
    else
    {
    	// Create group node.
    	RefPtr<GroupNode> groupNode = nodeSort.CreateGroupNode();

    	// Cache group node.
    	mNodeSortMap.insert( make_pair<>( nodeSort, groupNode ) );

    	// Attach node to branch.
    	AttachNode( node, groupNode );

        // Attach branch to root.
        AttachNode( groupNode.get(), mBaseRootNode.get() );
    }
}

/*****************************************************************************
 * Create a branch of the scene-graph at group node having special graphical-state.
 * The regular AttachNode() places nodes under a group with default graphical state.
 *****************************************************************************/
void
BaseSceneGraph::AttachBranchNode( RefPtr<GroupNode> groupNode )
{
	AttachNode( groupNode.get(), mBaseRootNode.get() );
}

/*****************************************************************************
 * Similar to AttachObject() but overrides Object::GetNodeSort().
 *****************************************************************************/
void
BaseSceneGraph::AttachObject( shptr<Object> object, const NodeSort& nodeSort )
{
CHECK_TYPESIG(&nodeSort,TYPESIG_NODE_SORT);

	// Attach Object's node.
	AttachNode( object->GetRootNode(), nodeSort );

	// AttachObject() implies that SceneGraph holds a reference to Object.
	// The C++ core needs a container of Objects (so this is it).
	mObjects.insert( object );
}

/*****************************************************************************
 * Detach node.
 * Ok to call if node is already detached.
 *****************************************************************************/
void
BaseSceneGraph::DetachNode( RefPtr<Node> node )
{
    // A node should be attached to 0 or 1 group nodes.
    const uint parentCnt = node->getNumParents();
    if ( parentCnt == 1 )
    {
    	// Has one parent which should be an osg::Group.
        RefPtr<osg::Group> parent = node->getParent( 0 );  // idx=0
    #if DEBUG
        ASSERT( parent->asGroup() != NULL );            // should be a Group or derivative
        ASSERT( parent->removeChild( node.get() ) );    // remove() should return true
    #else
        parent->removeChild( node.get() );
    #endif
    }
    else // parentCnt != 1
    {
        // Multiple parents shouldn't occur and isn't handled.
        // Objects may have shared nodes for 3D models but
        // those should be underneath unique group nodes (osg::MatrixTransform).
        ASSERT( parentCnt == 0 );
    }

    ASSERT( node->getNumParents() == 0 );  // should have zero parents
}

/*****************************************************************************
 * Detach user-defined branch node.
 * Ok to call if node is already detached.
 *****************************************************************************/
void
BaseSceneGraph::DetachBranchNode( RefPtr<GroupNode> branchNode )
{
	DetachNode( branchNode.get() );
}

/*****************************************************************************
 * Detach Object.
 * Ok to call if Object is already detached.
 *****************************************************************************/
void
BaseSceneGraph::DetachObject( shptr<Object> object )
{
	// Erasing is sufficient because ~Object() will detach its root node.
	mObjects.erase( object );
}

/*****************************************************************************
 * Run OSG optimizer on a node and its children.
 *****************************************************************************/
void
BaseSceneGraph::Optimize( RefPtr<Node> node, const NodeSort& nodeSort )
{
CHECK_TYPESIG(&nodeSort,TYPESIG_NODE_SORT);

	if ( GRAPH_CONF.mEnableOsgOptimizer )
	{
	    osgUtil::Optimizer optimizer;
	    optimizer.optimize( node.get(), nodeSort.GetOptimizations().mVal );
	}
}

/*****************************************************************************
 * Listen for the DestroyObject event.
 *****************************************************************************/
void
BaseSceneGraph::DestroyObjectListener::operator()( shptr<Object> object )
{
CHECK_TYPESIG(object,TYPESIG_OBJECT);

	mBaseSceneGraph.DetachObject( object );
}

} // namespace graph
