/*****************************************************************************
 * $LastChangedDate: 2009-05-21 11:41:41 -0500 (Thu, 21 May 2009) $
 * @file
 * @author	Jim E. Brooks  http://www.palomino3d.org
 * @brief	SceneGraph class.
 *//*
 * LEGAL:	COPYRIGHT (C) 2009 JIM E. BROOKS
 *          THIS SOURCE CODE IS RELEASED UNDER THE TERMS
 *          OF THE GNU GENERAL PUBLIC LICENSE VERSION 2 (GPL 2).
 *****************************************************************************/

#define GRAPH_SCENE_GRAPH_CC 1
#include "base/module.hh"
#include "base/conf.hh"
using namespace base;
#include "gfx/module.hh"
#include "gfx/rgba.hh"
#include "gfx/osg.hh"
using namespace gfx;
#include "object/module.hh"
using namespace object;
#include "graph/module.hh"
#include "graph/osg.hh"
#include "graph/conf.hh"
#include "graph/base_scene_graph.hh"
#include "graph/scene_graph.hh"
#include "graph/shadow_scene_graph.hh"

#include <osg/BlendFunc>
#include <osgUtil/Optimizer>
#include <osgDB/WriteFile>

namespace graph {

// True if node casts/receives shadows.
INLINE bool IF_SHADOW_NODE_SORT( const NodeSort& nodeSort )
{
    return (nodeSort.GetAttribs().mVal)
         & (NodeSort::Attribs::SHADOW_CASTER | NodeSort::Attribs::SHADOW_RECEIVER);
}

////////////////////////////////////////////////////////////////////////////////
//////////////////////////////  SceneGraph  ////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

SceneGraph* SceneGraph::msInstance;

/*****************************************************************************
 * ctor/dtor.
 *****************************************************************************/
SceneGraph::SceneGraph( void )
:   BaseSceneGraph(),
    mClearNode(new osg::ClearNode),  // root node
    mShadowSceneGraph(new ShadowSceneGraph(mClearNode.get()))
{
    // Set root node to an osg::ClearNode to change background color.
    BaseSceneGraph::SetRootNode( mClearNode.get() );

	// Set background color.
	SetBackgroundColor( GRAPH_CONF.mBackgroundColor );

    // Attach the shadowed scene-graph to this scene-graph.
    BaseSceneGraph::AttachNode( mShadowSceneGraph->GetRootNode().get(), NodeSort() );
}

SceneGraph::~SceneGraph()
{
	// NOP
}

/*****************************************************************************
 * Attach node to scene-graph under branch node with special graphical-state.
 *****************************************************************************/
void
SceneGraph::AttachNode( RefPtr<Node> node, const NodeSort& nodeSort )
{
CHECK_TYPESIG(&nodeSort,TYPESIG_NODE_SORT);

  	// Attach to either regular or shadowed scene-graph.
    if ( UX( IF_SHADOW_NODE_SORT(nodeSort) ) )
    {
        mShadowSceneGraph->AttachNode( node, nodeSort );
    }
    else
    {
        BaseSceneGraph::AttachNode( node, nodeSort );
    }
}

// Attach node with default graphical state.
void
SceneGraph::AttachNode( RefPtr<Node> node )
{
    AttachNode( node, NodeSort() );
}

/*****************************************************************************
 * Attach the node of an Object.
 * Implies that SceneGraph holds a reference to Object.
 * NodeSort is a hint for state-sorting.
 *****************************************************************************/
void
SceneGraph::AttachObject( shptr<Object> object, const NodeSort& nodeSort )
{
CHECK_TYPESIG(&nodeSort,TYPESIG_NODE_SORT);

  	// Attach to either regular or shadowed scene-graph.
    if ( UX( IF_SHADOW_NODE_SORT(nodeSort) ) )
    {
        mShadowSceneGraph->AttachObject( object, nodeSort );
    }
    else
    {
        BaseSceneGraph::AttachObject( object, nodeSort );
    }
}

void
SceneGraph::AttachObject( shptr<Object> object )
{
    AttachObject( object, object->GetNodeSort() );
}

/*****************************************************************************
 * Detach node.
 * Ok to call if node is already detached.
 *****************************************************************************/
void
SceneGraph::DetachNode( RefPtr<Node> node )
{
    BaseSceneGraph::DetachNode( node );
    mShadowSceneGraph->DetachNode( node );
}

/*****************************************************************************
 * Detach Object.
 * Ok to call if Object is already detached.
 *****************************************************************************/
void
SceneGraph::DetachObject( shptr<Object> object )
{
	BaseSceneGraph::DetachObject( object );
    mShadowSceneGraph->DetachObject( object );
}

/*****************************************************************************
 * Optimize entire scene-graph.
 *****************************************************************************/
void
SceneGraph::Optimize( void )
{
#if 0
	// This is both redundant and wrong.
	// Applying default optimizations would undo and break special models
	// that require specific NodeSort::Optimizations values.
	// However, this method remains as part of the gfxsys-neutral interface.
	Optimize( GetRootNode().get(), NodeSort() );
#endif
}

/*****************************************************************************
 * Set the background color.
 *****************************************************************************/
void
SceneGraph::SetBackgroundColor( const RGBA color )
{
	mClearNode->setClearColor( RGBA2VEC4F(color) );
}

/*****************************************************************************
 * Dump scene graph to file.
 *****************************************************************************/
void
SceneGraph::SaveToFile( const string& filename )
{
	COUT << "Saving scene-graph to file " << filename << "...";
	if ( osgDB::writeNodeFile( *GetRootNode(), filename ) )
		COUT << "ok." << endl;
	else
		CERROR << "ERROR" << endl;
}

} // namespace graph
