/*****************************************************************************
 * $LastChangedDate: 2009-06-15 08:35:22 -0500 (Mon, 15 Jun 2009) $
 * @file
 * @author	Jim E. Brooks  http://www.palomino3d.org
 * @brief	Sky model (Singleton).
 *//*
 * LEGAL:   COPYRIGHT (C) 2008 JIM E. BROOKS
 *          THIS SOURCE CODE IS RELEASED UNDER THE TERMS
 *          OF THE GNU GENERAL PUBLIC LICENSE VERSION 2 (GPL 2).
 *****************************************************************************/

#define SKY_SKY_CC 1
#include "base/module.hh"
using namespace base;
#include "gfx/module.hh"
#include "gfx/osg.hh"
using namespace gfx;
#include "glue/module.hh"
#include "glue/sim_time.hh"
using namespace glue;
#include "graph/module.hh"
#include "graph/scene_graph.hh"
using namespace graph;
#include "shader/module.hh"
#include "shader/light.hh"
using namespace shader;
#include "view/module.hh"
#include "view/window.hh"
using namespace view;
#include "world/module.hh"
using namespace world;
#include "sky/module.hh"
#include "sky/sky.hh"
#include "sky/conf.hh"
#include "sky/defs.hh"
#include "sky/sky_dome.hh"
#include "sky/sky_dome_painter.hh"
#include "sky/clouds.hh"
#include "sky/sun.hh"
#include "sky/moon.hh"
#include "sky/stars.hh"

namespace sky {

////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////  Sky  ///////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

Sky* Sky::msInstance;

/*****************************************************************************
 * ctor/dtor.
 *****************************************************************************/
Sky::Sky( void )
:	mSkyDome(new SkyDome( GET_WORLD().GetPosition(),
                          GET_WORLD().GetRadius(),
                          SKY_CONF.mSkyColor,
                          SKY_CONF.mSkyDomeSliceCnt ) ),  // -- ORDER-DEPENDENT --
	mCloudParticleSystem0(new CloudParticleSystem(*mSkyDome,GetVolume(),4,"textures/cloud0.png")),
	mCloudParticleSystem1(new CloudParticleSystem(*mSkyDome,GetVolume(),4,"textures/cloud1.png")),
	mCloudParticleSystem2(new CloudParticleSystem(*mSkyDome,GetVolume(),4,"textures/cloud2.png")),
	mCloudParticleSystem3(new CloudParticleSystem(*mSkyDome,GetVolume(),4,"textures/cloud3.png")),
	mSun(new Sun( sky::defs::DEFAULT_SUN_LAT, sky::defs::DEFAULT_SUN_LON ) ),
	mMoon(new Moon( *mSkyDome,
                    sky::defs::DEFAULT_MOON_LAT,
                    sky::defs::DEFAULT_MOON_LON ) ),
	mStars(new Stars( GET_WORLD().GetPosition(),
                      SKY_CONF.mStarCnt,
                      *mSkyDome ) )
{
CHECK_TYPESIG(mSkyDome,TYPESIG_SKY_DOME);
ASSERT( GET_WORLD().GetRadius() > 0.0f );
ASSERT( SKY_CONF.mSkyDomeSliceCnt > 0 );

	// NOP.  The ctor creates the Objects.
	// Later, Sky::Start() will attach them to the scene graph.

#if DEBUG
{
	// Self-check.
	pair<Degree,Degree> latLon;
	latLon = ComputeLatLonFromHour(  0.0f ); ASSERT( FP_EQ<fp>( latLon.first, 180.0f ) ); ASSERT( FP_EQ<fp>( latLon.second, 0.0f ) );
	latLon = ComputeLatLonFromHour(  6.0f ); ASSERT( FP_EQ<fp>( latLon.first,  90.0f ) ); ASSERT( FP_EQ<fp>( latLon.second, 0.0f ) );
	latLon = ComputeLatLonFromHour( 12.0f ); ASSERT( FP_EQ<fp>( latLon.first,   0.0f ) ); // lon doesn't matter at poles
	latLon = ComputeLatLonFromHour( 18.0f ); ASSERT( FP_EQ<fp>( latLon.first,  90.0f ) ); ASSERT( FP_EQ<fp>( latLon.second, 180.0f ) );
}
#endif
}

Sky::~Sky()
{
	// NOP
}

/*****************************************************************************
 * Attach the sky nodes to the scene graph.
 *****************************************************************************/
void
Sky::Start( const fp hour )
{
CHECK_TYPESIG(mSkyDome,TYPESIG_SKY_DOME);

BEGIN_ONCE {

ASSERT_HOUR(hour);

	// Apply time-of-day to components of sky.
	SetHour( hour );

	// Attach the Object components of the sky to the scene graph
	// in appropriate state-sorted group nodes.
	if ( GET_WORLD().IfFlat() )
	{
		GET_SCENE_GRAPH().AttachObject( mSkyDome.PTR() );
       	GET_SCENE_GRAPH().AttachObject( mMoon.PTR() );
    }
  	GET_SCENE_GRAPH().AttachObject( mStars.PTR() );

	// Initialize lighting and fog.
	InitLightingAndFog( hour );

} END_ONCE
}

/*****************************************************************************
 * Initialize lighting and fog.
 *****************************************************************************/
void
Sky::InitLightingAndFog( const fp hour )
{
    // (superceded by shaders)
}

/*****************************************************************************
 * Update sky for time-of-day.
 *****************************************************************************/
void
Sky::SetHour( const fp hour )
{
ASSERT_HOUR(hour);

    // Change intensity/color of light.
    RGBA baseAmbient( 0.0f, 0.0f, 0.0f, 0.0f );
    fp brightness = 1.0f;
    if ( SimTime::IfNightTime( hour ) )
    {
        baseAmbient.Set( 0.0f, 0.0f, 0.3f, 0.0f );
        brightness = 0.3f;
    }
    else if ( SimTime::IfMorningTime( hour ) or SimTime::IfEveningTime( hour ) )
    {
        baseAmbient.Set( 0.05f, 0.05f, 0.15f, 0.0f );
        brightness = 0.7f;
    }
    GET_LIGHT().SetLightBaseAmbient( baseAmbient );
    GET_LIGHT().SetLightBrightness( brightness );

	// Change position of sun.
#if 0
	mSun->SetLatLon( ComputeLatLonFromHour( hour ) );
#else
    // Make sun visible thru particular random arrangement of clouds.
	mSun->SetLatLon( make_pair<>( Degree(82.0f), Degree(40.0f) ) );
#endif

	// Re-paint the sky dome.
	SkyDomePainter::Paint( *mSkyDome, hour );

    // For OSSIM, space is black.
	if ( not GET_WORLD().IfFlat() )
    {
        GET_SCENE_GRAPH().SetBackgroundColor( RGBA( 0.0f, 0.0f, 0.15f, 1.0f ) );
    }
}

/*****************************************************************************
 * Compute (lat,lon) of sun from time-of-day.
 *****************************************************************************/
pair<Degree,Degree>
Sky::ComputeLatLonFromHour( const fp hour )
{
ASSERT_HOUR(hour);

	// hour  0: (lat=180,lon=0)
	// hour  6: (lat= 90,lon=0)
	// hour 12: (lat=  0,lon=0 or 180)
	// hour 18: (lat= 90,lon=180)
	// hour 24: (lat=180,lon=0 or 180)
	const Degree lat = ABS(hour - 12.0f) * (180.0f / 12.0f);
	const Degree lon = (hour < 12.0f) ? 0.0f : 180.0f;
	ASSERT_LAT_LON_DEG( lat, lon );
	return make_pair<Degree,Degree>( lat, lon );
}

/*****************************************************************************
 * @return Volume of sky in 3D space.
 *****************************************************************************/
BoxVolume
Sky::GetVolume( void )
{
	return GET_WORLD().GetVolume();
}

/*****************************************************************************
 * Accessors.
 *****************************************************************************/
SkyDome&
Sky::GetSkyDome( void )
{
CHECK_TYPESIG(mSkyDome,TYPESIG_SKY_DOME);

	return *mSkyDome;
}

Sun&
Sky::GetSun( void )
{
	return *mSun;
}

} // namespace sky
