/*****************************************************************************
 * $LastChangedDate: 2009-06-15 11:51:43 -0500 (Mon, 15 Jun 2009) $
 * @file
 * @author	Jim E. Brooks  http://www.palomino3d.org
 * @brief	Module configuration.
 *//*
 * LEGAL:	COPYRIGHT (C) 2007 JIM E. BROOKS
 *          THIS SOURCE CODE IS RELEASED UNDER THE TERMS
 *          OF THE GNU GENERAL PUBLIC LICENSE VERSION 2 (GPL 2).
 *****************************************************************************/

#define WORLD_CONF_CC 1
#include "base/module.hh"
#include "base/conf.hh"
#include "base/misc.hh"
using namespace base;
#include "math/module.hh"
#include "math/funcs.hh"
using namespace math;
#include "gfx/module.hh"
#include "gfx/rgba.hh"
using namespace gfx;
#include "control/module.hh"
using namespace control;
#include "world/module.hh"
#include "world/conf.hh"

namespace world {

////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////  Conf  ///////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

/*****************************************************************************
 * ctor/dtor.
 *****************************************************************************/
Conf::Conf( void )
:	mValid(false),
	// cmd-line args
	mFlat(true),
	mUseOssim(false),
	mEnableOsgOptimizer(true),
	// .conf file
	// World:
	mObjectScale(1.0),
	mObjectScale3(1.0,1.0,1.0),
	mFlatTerrainSceneGraphFile(""),
	mFlatTerrainSceneGraphScale(1.0f),
	mFlatNW(Degree(0),Degree(0),Meter(0)),
	mFlatSE(Degree(0),Degree(0),Meter(0)),
	mFlatMeterOffset(Meter(0),Meter(0),Meter(0)),
	mFlatMeterSimRatio(1.0f)
{
}

/*****************************************************************************
 * Initialize/read configuration.
 *****************************************************************************/
void
Conf::Init( int argc, char** argv, const string& confTxtFilename )
{

BEGIN_ONCE {

	mValid = true;  // assign this first as *_CONF may expand to GetConf() which asserts it

	ParseCmdlineArgs( argc, argv );   // cmdline args parsed prior to reading .conf file
	ReadConfTxt( confTxtFilename );   // read from textual .conf file
	ParseCmdlineArgs2( argc, argv );  // cmdline args that override .conf file

} END_ONCE

}

/*****************************************************************************
 * Parse cmdline args.
 *****************************************************************************/
void
Conf::ParseCmdlineArgs( int argc, char** argv )
{
	for ( int i = 1; i < argc; ++i )
	{
		// -llh/lla is handled by Lua scripts.

		if ( STREQ( argv[i], "-ossim" ) )
		{
        #if COMPILE_OSSIM
			mFlat = false;
			mUseOssim = true;
			CLOG << "Enabled OSSIM." << endl;
		#else
            CERROR << "ERROR: OSSIM support isn't compiled." << endl;
            EXIT( 1 );
        #endif
			CmdargDone( i );
		}
		else if ( STREQ( argv[i], "-no-ossim" ) )
		{
			mFlat = true;
			mUseOssim = false;
			CmdargDone( i );
			CLOG << "Disabled OSSIM." << endl;
		}
		else if ( STREQ( argv[i], "-no-osg-opt" ) )
		{
			mEnableOsgOptimizer = false;
			CmdargDone( i );
			CLOG << "Disabled OSG optimizer." << endl;
		}
	}
}

void
Conf::ParseCmdlineArgs2( int argc, char** argv )
{
	// NOP
}

/*****************************************************************************
 * Read from .conf file.
 *****************************************************************************/
void
Conf::ReadConfTxt( const string& confTxtFilename )
{
	try
	{
		ModuleConfTxt confTxt;
		confTxt.Read( string(base::GetDataDir()) + string(confTxtFilename) );
	}
	catch (...)
	{
		CERROR << "ERROR: cannot read " << confTxtFilename << endl;
		EXIT( 1 );
	}
}

void
Conf::ModuleConfTxt::ProcessField( string& field, ifstream& conf )
{
	bool error = false;

	//---------------------  flat/sphere world  --------------------------------
	if ( field == "FlatObjectScale" and not WORLD_CONF.mUseOssim )
	{
		conf >> WORLD_CONF.mObjectScale;
		WORLD_CONF.mObjectScale3[XX] = \
		WORLD_CONF.mObjectScale3[YY] = \
		WORLD_CONF.mObjectScale3[ZZ] = WORLD_CONF.mObjectScale;
	}
	else if ( field == "SphereObjectScale" and WORLD_CONF.mUseOssim )
	{
		conf >> WORLD_CONF.mObjectScale;
		WORLD_CONF.mObjectScale3[XX] = \
		WORLD_CONF.mObjectScale3[YY] = \
		WORLD_CONF.mObjectScale3[ZZ] = WORLD_CONF.mObjectScale;
	}
	else if ( field == "FlatTerrainSceneGraphFile" )
	{
		conf >> WORLD_CONF.mFlatTerrainSceneGraphFile;
	}
	else if ( field == "FlatTerrainSceneGraphScale" )
	{
		conf >> WORLD_CONF.mFlatTerrainSceneGraphScale;
	}
	else if ( field == "FlatMeterSimRatio" )
	{
		conf >> WORLD_CONF.mFlatMeterSimRatio;
	}
	else if ( field == "FlatNW" )
	{
		fp f;
		conf >> f; WORLD_CONF.mFlatNW.mLat = Degree(f);  // SphereVertex
		conf >> f; WORLD_CONF.mFlatNW.mLon = Degree(f);
		conf >> f; WORLD_CONF.mFlatNW.mAlt = Meter(f);
	}
	else if ( field == "FlatSE" )
	{
		fp f;
		conf >> f; WORLD_CONF.mFlatSE.mLat = Degree(f);  // SphereVertex
		conf >> f; WORLD_CONF.mFlatSE.mLon = Degree(f);
		conf >> f; WORLD_CONF.mFlatSE.mAlt = Meter(f);
	}
	else if ( field == "FlatMeterOffset" )
	{
		fp f;
		conf >> f; WORLD_CONF.mFlatMeterOffset[XX] = Meter(f);
		conf >> f; WORLD_CONF.mFlatMeterOffset[YY] = Meter(f);
		conf >> f; WORLD_CONF.mFlatMeterOffset[ZZ] = Meter(f);
	}

    // Errors?
	if ( error )
	{
		CERROR << "ERROR[conf]: parameter for " << field << " is invalid." << endl;
		EXIT( 1 );
	}
}

/*****************************************************************************
 * Print Conf configuration.
 *****************************************************************************/
ostream&
operator<<( ostream& strm, const Conf& conf )
{
	return strm;
}

} // namespace world
