/*****************************************************************************
 * $LastChangedDate: 2009-06-15 11:51:43 -0500 (Mon, 15 Jun 2009) $
 * @file
 * @author	Jim E. Brooks  http://www.palomino3d.org
 * @brief	World class.
 *//*
 * LEGAL:   COPYRIGHT (C) 2007 JIM E. BROOKS
 *          THIS SOURCE CODE IS RELEASED UNDER THE TERMS
 *          OF THE GNU GENERAL PUBLIC LICENSE VERSION 2 (GPL 2).
 *****************************************************************************/

#define WORLD_WORLD_CC 1
#define NEED_PLANET_CLASS 1
#include "base/module.hh"
#include "base/conf.hh"
using namespace base;
#include "math/module.hh"
#include "math/matrix.hh"
#include "math/matrix_funcs.hh"
#include "math/funcs_trig.hh"
using namespace math;
#include "gfx/module.hh"
#include "gfx/osg.hh"
using namespace gfx;
#include "object/module.hh"
#include "object/events.hh"
using namespace object;
#include "graph/module.hh"
#include "graph/scene_graph.hh"
using namespace graph;
#include "view/module.hh"
using namespace view;
#include "physics/module.hh"
#include "physics/types.hh"
#include "physics/physics_control.hh"
using namespace physics;
#include "sky/module.hh"
#include "sky/conf.hh"
using namespace sky;
#include "world/module.hh"
#include "world/defs.hh"
#include "world/conf.hh"
#include "world/planet.hh"

namespace world {

INTERN const Meter APPROX_LOW_ALTITUDE_METERS( 100.0f );

////////////////////////////////////////////////////////////////////////////////
//////////////////////////////  World::Make  ///////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

/*****************************************************************************
 * Called by world::Init() [world/module.cc].
 *****************************************************************************/
void
World::Make( void )
{
BEGIN_ONCE {

	// Set the background color.
	if ( WORLD_CONF.mFlat )
	    GET_SCENE_GRAPH().SetBackgroundColor( RGBA(0.1f,0.1f,0.5f,1.0f) );  // dark blue
	else  // [2008/04] sky not fully implemented for OSSIM
		GET_SCENE_GRAPH().SetBackgroundColor( SKY_CONF.mSkyColor );

	// Create and attach Planet.  OSSIM is too dark with lighting.
	shptr<Object> planet = GET_PLANET().GetRootObject();
	planet->SetCollidable( true );  // planet is a collidable object
	GET_SCENE_GRAPH().AttachNode( planet->GetRootNode(),
                                  WORLD_CONF.mFlat ? NodeSort()
                                                   : NodeSort( NodeSort::Attribs( NodeSort::Attribs::NO_SHADER ) ) );
} END_ONCE
}

////////////////////////////////////////////////////////////////////////////////
////////////////////////////  World : info  ////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

/*****************************************************************************
 * Get radius of World/Planet.
 *****************************************************************************/
fp
World::GetRadius( void )
{
	return GET_PLANET().RadiusGraphical();
}

/*****************************************************************************
 * @return Volume of world in 3D space.
 *****************************************************************************/
BoxVolume
World::GetVolume( void )
{
	return GET_PLANET().GetVolume();
}

/*****************************************************************************
 * If the World is flat ;-)
 * This is queried since FlatPlanet is deficient and some configuration
 * is only applicable to FlatPlanet.
 *****************************************************************************/
bool
World::IfFlat( void )
{
	return GET_PLANET().IfFlat();
}

/*****************************************************************************
 * @return True if position is approx at low altitude.
 *****************************************************************************/
bool
World::IfLowAltitude( const WorldVertex& pos )
{
    return world::conv::WorldVertex2SphereVertex(pos).mAlt < APPROX_LOW_ALTITUDE_METERS;
}

////////////////////////////////////////////////////////////////////////////////
/////////////////////////////  World orientation  //////////////////////////////
////////////////////////////////////////////////////////////////////////////////

/*****************************************************************************
 * Compute a normalized "up vector" extending from a point on the world.
 * A "down vector" can be computed by multipling by -1.0f.
 *****************************************************************************/
WorldVertex
World::ComputeUpVector( const SphereVertex& sv )
{
	return GET_PLANET().ComputeUpVector( sv );
}

/*****************************************************************************
 * Compute an "up vector" where an Object is at.
 *****************************************************************************/
WorldVertex
World::ComputeUpVector( Object& object )
{
CHECK_TYPESIG(&object,TYPESIG_OBJECT);

	return ComputeUpVector( world::conv::WorldVertex2SphereVertex( object.GetPosition() ) );
}

/*****************************************************************************
 * Orient an Object to be tangent to a point on the world (in case world is spherical).
 *****************************************************************************/
void
World::RotateTangential( Object& object, const SphereVertex& sv )
{
	if ( not GET_PLANET().IfFlat() )
	{
		Matrix m = GetTangentMatrix( sv );
		m.setTrans( object.GetPosition() );  // keep Object's position
		MatrixRotateLocal( m, XX, Deg2Rad(90.0f) );  // otherwise Object points up (normal to surface)
		object.SetMatrix( m );
	}
	// else flat so NOP
}

/*****************************************************************************
 * Compute a matrix that is tangent to a point on the world.
 *****************************************************************************/
const Matrix&
World::GetTangentMatrix( const SphereVertex& sv )
{
	return GET_PLANET().GetTangentMatrix( sv );
}

} // namespace world
